# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Test whether or not a node is compatible with MAAS."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

__metaclass__ = type
__all__ = [
    "main",
    ]

import logging
import sys
import unittest

from maastest.kvmfixture import KVMFixture
import testresources
import testtools
from testtools.testcase import gather_details


class TestMAAS(testtools.TestCase, testresources.ResourcedTestCase):

    @classmethod
    def setUpClass(cls):
        # TODO: series and architecture should be script parameters.
        cls.machine = KVMFixture(series='saucy', architecture='amd64')
        cls.machine.setUp()
        super(TestMAAS, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        cls.machine.cleanUp()
        super(TestMAAS, cls).tearDownClass()

    @classmethod
    def configure(cls, interface):
        cls.direct_interface = interface

    def shortDescription(self):
        """Overridable from `unittest`: Describe the ongoing test.

        By default this shows the first line of the test's docstring (if
        available), but this `TestCase` is basically our UI so we want full,
        user-friendly docstrings.
        """
        # Cribbed from unittest2.TestCase.  The difference is that we return
        # the whole docstring, instead of just the first line.
        # Future maintainers beware: the docstring may also be None.
        return self._testMethodDoc

    # XXX 2013-11-07 gmb bug=897190:
    #     This exists because of the above bug in testtools, which means
    #     that fixture details don't get included with the TestCase
    #     details on cleanUp().
    def unify_details(self):
        """Merge the details of the fixture into the details of the TestMAAS.
        """
        gather_details(self.machine.getDetails(), self.getDetails())

    def setUp(self):
        super(TestMAAS, self).setUp()
        self.addCleanup(self.unify_details)

    def test_install_maas_package(self):
        """Installing MAAS package"""
        # Install the English language pack first.  If this is not done before
        # postgres is installed, on some systems it won't be able to set up its
        # main cluster.
        # TODO: Is there no way to ensure this through a dependency?
        self.machine.run_command([
            'sudo', 'DEBIAN_FRONTEND=noninteractive',
            'apt-get', 'install', '-y', 'language-pack-en',
            ])
        # Now we can install maas (which also installs postgres).
        retcode, _, stderr = self.machine.run_command([
            'sudo', 'DEBIAN_FRONTEND=noninteractive',
            'apt-get', 'install', '-y', 'maas', 'maas-dhcp', 'maas-dns',
            ])
        self.assertEqual(0, retcode)

# TODO:
# Configure the main cluster's DHCP using the IP range provided by the fixture.
#    def test_configure_cluster_dhcp(self):
#        pass
#        # Configure the main cluster's dhcp using the IP range:
#        # [
#        #     self.machine.direct_first_available_ip,
#        #     "%s" % netaddr.IPAddress(self.machine.direct_network.last)
#        # ]


def main(argv=None):
    # TODO: control the order in which the tests are run.  By default
    # the tests are run in alphabetical order and we should use the order
    # in which the tests are declared in the source code instead.
    from maastest import parser
    maas_test_parser = parser.prepare_parser()
    args = maas_test_parser.parse_args()
    output_stream = sys.stderr
    logging.basicConfig(
        level=logging.DEBUG, stream=output_stream, format="%(message)s")

    class ConfiguredTestMAAS(TestMAAS):
        """A configured version of TestMAAS.

        ConfiguredTestMAAS is a TestMAAS use to configure it by calling
        cls.configure.  We need to configure the class itself and not
        the instance because the KVMFixture TestMAAS instanciates and needs
        to configure is created at the class level.
        """
    ConfiguredTestMAAS.configure(args.interface)
    suite = unittest.TestLoader().loadTestsFromTestCase(ConfiguredTestMAAS)
    runner = unittest.TextTestRunner(
        stream=output_stream, verbosity=2, descriptions=False)
    runner.run(suite)


if __name__ == '__main__':
    main()
