# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Helpers for setting up MAAS in a virtual machine."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

str = None

__metaclass__ = type
__all__ = [
    'configure_maas',
    'create_maas_admin',
    'determine_vm_series',
    'determine_vm_architecture',
    'import_maas_images',
    'query_api_key',
    ]

import platform
from random import choice
from string import letters


MAAS_ADMIN_USER = 'admin'


class UnknownCPUArchitecture(RuntimeError):
    """Could not determine CPU architecture."""


def determine_vm_architecture():
    """Figure out CPU architecture for the virtual machine that will run MAAS.

    :return: Architecture string, e.g. 'i386'.
    :raise UnknownCPUArchitecture: if the CPU architecture could not be
        determined.
    """
    # Ubuntu sometimes uses different architecture names than we get from
    # the platform module.  This dict translates for those cases.
    ubuntu_names = {
        'i686': 'i386',
        'x86_64': 'amd64',
    }
    raw_arch = platform.machine()
    if raw_arch == '':
        raise UnknownCPUArchitecture(
            "Could not determine system CPU architecture.  "
            "Please report a bug against http://launchpad.net/maas-test "
            "with this system's details.  To work around the problem, "
            "try the --vm-arch option.")
    return ubuntu_names.get(raw_arch, raw_arch)


def determine_vm_series():
    """Figure out the Ubuntu release series to run MAAS on.

    :return: Series codename, e.g. "precise" for 12.04 Precise Pangolin.
    """
    distro, version, series = platform.linux_distribution()
    return series


def query_api_key(kvm_fixture, username):
    """Return the API key for the given MAAS user."""
    # The "apikey" command prints the user's API key to stdout.
    return_code, stdout, stderr = kvm_fixture.run_command([
        'sudo', 'maas', 'apikey',
        '--username=%s' % username,
        ], check_call=True)
    return stdout.strip()


def create_maas_admin(kvm_fixture):
    """Create a MAAS admin user.

    Invoke this after MAAS has been installed on the virtual machine.

    :return: User name and password for the new admin.
    """
    username = MAAS_ADMIN_USER
    password = ''.join(choice(letters) for counter in range(8))
    # Email address really does not matter.  MAAS never sends anything there.
    # Its hostname doesn't have to resolve, although it has to look like a
    # FQDN.
    email = 'root@localhost.local'

    return_code, _, stderr = kvm_fixture.run_command([
        'sudo', 'maas', 'createadmin',
        '--username=%s' % username,
        '--password=%s' % password,
        '--email=%s' % email,
        ], check_call=True)

    return username, password


def import_maas_images(kvm_fixture, admin_user, api_key):
    """Import boot images into the MAAS instance.

    This download gigabytes of data to the virtual machine's disk.

    :param kvm_fixture: `KVMFixture` that controls the virtual machine.
    :param admin_user: Login name for the MAAS admin user.
    :param api_key: Admin user's API key.
    """
    # TODO: Actually implement.


def configure_maas(kvm_fixture):
    """Configure the MAAS running inside the virtual machine.

    MAAS must already have been installed.  This function creates an admin user
    and imports boot images.

    :param kvm_fixture: `KVMFixture` that controls the virtual machine.
    :return: Username and API key for the MAAS admin user.
    """
    admin_user, password = create_maas_admin(kvm_fixture)
    api_key = query_api_key(kvm_fixture, admin_user)
    import_maas_images(kvm_fixture, admin_user, api_key)
