# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the "prepare" command."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

str = None

__metaclass__ = type
__all__ = []

import platform

from maastest import prepare
from maastest.kvmfixture import KVMFixture
from mock import MagicMock
from testtools import TestCase


class TestPrepare(TestCase):

    def make_kvm_fixture(self):
        """Create a `KVMFixture` with arbitrary series/architecture.

        Its `run_command` method will be replaced by a `Mock`, which returns
        success by default.
        """
        fixture = KVMFixture('droopy', 'i860')
        fixture.run_command = MagicMock(
            return_value=(0, 'stdout output', 'stderr output'))
        return fixture

    def test_determine_vm_architecture_gets_system_arch(self):
        # Additions may become necessary.
        supported_architectures = {
            'amd64',
            'armhf',
            'i386',
        }
        self.assertIn(
            prepare.determine_vm_architecture(),
            supported_architectures)

    def test_determine_vm_architecture_returns_ubuntu_arch_name(self):
        # Architecture names according to Python's platform module, and their
        # Ubuntu equivalents.  Unknown names are passed on unchanged.
        arch_names = {
            'i386': 'i386',
            'i686': 'i386',
            'x86_64': 'amd64',
            'armhf': 'armhf',
            'power': 'power',
            'sparc64': 'sparc64',
            'm68k': 'm68k',
        }
        # See what architectures determine_vm_architecture() reports for these
        # systems.
        translations = {}
        machine_func = MagicMock()
        self.patch(platform, 'machine', machine_func)
        for arch in arch_names.keys():
            machine_func.return_value = arch
            translations[arch] = prepare.determine_vm_architecture()

        self.assertEqual(arch_names, translations)

    def test_determine_vm_architecture_reports_failure(self):
        # platform.machine() returns an empty string when it can't determine
        # the architecture.  This test simulates that scenario.
        # In python 3 it returns a unicode string, in python 2 a string of
        # bytes.
        return_type = platform.machine().__class__
        self.patch(platform, 'machine', lambda: return_type())

        self.assertRaises(
            prepare.UnknownCPUArchitecture,
            prepare.determine_vm_architecture)

    def test_determine_vm_series_returns_system_series(self):
        # We don't know the names of future series this test might run on, but
        # it'll be a lower-case name.  The supported releases start with
        # Precise, and for the forseeable future, progress through the
        # alphabet from there.
        self.assertRegexpMatches(
            prepare.determine_vm_series(),
            '[p-z][a-z]+$')

    def test_determine_vm_series_uses_distro_information(self):
        series = 'nerdy'
        distro = ('Ubuntu', '1.01', series)
        self.patch(platform, 'linux_distribution', lambda: distro)

        self.assertEqual(series, prepare.determine_vm_series())

    def test_create_maas_admin_creates_admin(self):
        fixture = self.make_kvm_fixture()

        username, password = prepare.create_maas_admin(fixture)

        create_call = fixture.run_command.call_args_list[0]
        args, kwargs = create_call
        command_line = args[0]
        self.assertEqual(
            ['sudo', 'maas', 'createadmin'],
            command_line[:3])
        self.assertIn('--username=%s' % username, command_line)
        self.assertIn('--password=%s' % password, command_line)

    def test_query_api_key_returns_api_key(self):
        username = "maas-user"
        actual_api_key = "API KEY"
        fixture = self.make_kvm_fixture()
        fixture.run_command.return_value = (0, actual_api_key, 'stderr output')

        api_key = prepare.query_api_key(fixture, username)

        args, kwargs = fixture.run_command.call_args_list[0]
        command_line = args[0]
        self.assertItemsEqual(
            ['sudo', 'maas', 'apikey', '--username=%s' % username],
            command_line)
        self.assertEqual(actual_api_key, api_key)
